/** Copyright (C) 2008 Philippe Dreuw
 * 
 * @file    ProgressIndicator.hh
 * @author  Philippe Dreuw <dreuw@cs.rwth-aachen.de>
 * @brief   Progress indicator which povides assurance to impacient users.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * $Id: ProgressIndicator.hh,v 1.3 2008-02-25 15:54:28 dreuw Exp $
 * 
 */
#ifndef CORE_PROGRESSINDICATOR_HH
#define CORE_PROGRESSINDICATOR_HH

#include <ctime>
#include <stdint.h>
#include <string>
#include <vector>

namespace Core {

    /** Provide assurance to impacient users.
     *
     * A progress indicators provides a sign of activity during
     * lengthy operations.  The only purpose is to reassure the user
     * that the machine is still working and will eventually complete.
     *
     * This class has been designed for minimal performance penalty.
     * Unlike simple-minded printf("%d\r", i) approaches, you may
     * safely use ProgressIndicator::notify() from the inner-most loop
     * of a highly compute intensive algorithm.
     *
     * Usage:
     *
     * Usually you will have some way to measure the amount of work
     * done. E.g. "bytes transferred", "time frames processed", "lines
     * read".  Indeally you know the total number of these "work
     * units" in advance, or have an upper bound which will be lowered
     * later on.
     * -# Create an instance of ProgressIndicator.
     * -# Call start().  If known, provide the total number units to
     * process.
     * -# Call notify() with the current number of units processed
     * each time you come around. If necessary you can also update the
     * total.  If you have no idea how much work has been done, just
     * call notify without arguments.
     * -# When your task is complete, call finish().  This will hide
     * the progress indicator.
     *
     * Limitations:
     * - Currently there can be at most one progress indicator active
     *   at any time.
     * - There are no provisions for thead safety.
     *
     * TODO:
     * - Should detect if cout is conneted to a TTY, and suppress
     *   messages otherwise.
     * - see Limitations
     */

    typedef int32_t     s32;
    typedef uint32_t    u32;
    typedef float       f32;

    class ProgressIndicator {
    public:
	enum Alignment { Left, Right };

    private:
	static const int defaultLength;
	static const int fd = 2;
	Alignment align_;
	std::string task_, unit_;
	u32 tick_, done_, total_, doneAtLastUpdate_, totalAtLastUpdate_;
	time_t startTime_;
	bool isVisible_;
	class DrawBuffer;
	DrawBuffer *draw;

	void hide();
	void print();
	void update();
	void updateSize();

	static ProgressIndicator *activeInstance;
	static void sigAlarmHandler(int);
	static void sigStopContinueHandler(int);
	static void sigWinchHandler(int);
	void activate();
	void deactivate();
	static const float updateInterval; // seconds
	bool isEnabled() const;
	bool isInForeground() const;
	int windowSize();
    public:
	/**
	 * @param task a terse description of the operation, which is
	 * displayed to the left of the progress indicator.
	 *
	 * @param unit a optional name of the unit progress is
	 * measured in (e.g. bytes, lines, states, nodes...)
	 **/
	ProgressIndicator(const std::string &task, const std::string &unit = "");
	~ProgressIndicator();

	/**
	 * Indicate start of the operation.
	 * @param total (estimated) amount of work to do.
	 */
	void start(u32 total = 0);

	/**
	 * Change task description.
	 * @param task a terse description of the operation, which is
	 * @param align indicate wether description should be left or right aligned
	 * displayed to the left of the progress indicator.
	 */
	void setTask(const std::string&, Alignment align = Left);

	/**
	 * Update estimate of total work.
	 */
	void setTotal(u32 total) {
	    if (total > total_) total_ = total;
	}

	/**
	 * Indicate the current state of operation.
	 * You may call notify() very freuqently, e.g. from your inner
	 * loop. The run-time penalty of notify() is very small, since
	 * actual updates of the indicator (involving output) are done
	 * only a few times per second.
	 */
	void notify(u32 done) {
	    if (done > done_) done_ = done;
	}

	/**
	 * Indicate the current state of operation and update estmate
	 * of total work.  This is similar to notify(done).
	 */
	void notify(u32 done, u32 total) {
	    setTotal(total);
	    notify(done);
	}

	/**
	 * Indicate that the operation is going on.
	 */
	void notify() {
	    notify(done_ + 1);
	}

	/**
	 * Indicate completion of the operation.
	 * @param hide, if true (default) the progress indicator is
	 * erased, otherwise it remains visible.
	 */
	void finish(bool hide = true);
    };

    class WithProgressIndicator {
    private:
	mutable ProgressIndicator progress_;
    public:
	WithProgressIndicator(const std::string &task = "", const std::string &unit = "") : progress_(task, unit) {}
	void setTask(const std::string &task, ProgressIndicator::Alignment align) const { progress_.setTask(task, align); }
	void start(u32 total = 0) const { progress_.start(total); }
	void setTotal(u32 total) const { progress_.setTotal(total); }
	void notify(u32 done) const { progress_.notify(done); }
	void notify(u32 done, u32 total) const { progress_.notify(done, total); }
	void notify() const { progress_.notify(); }
	void finish(bool hide = true) const { progress_.finish(hide); }
    };

    class WithoutProgressIndicator {
    public:
	WithoutProgressIndicator(const std::string &task = "", const std::string &unit = "") {}
	void setTask(const std::string &task, ProgressIndicator::Alignment align) const {}
	void start(u32 total = 0) const {}
	void setTotal(u32 total) const {}
	void notify(u32 done) const {}
	void notify(u32 done, u32 total) const {}
	void notify() const {}
	void finish(bool hide = true) const {}
    };

} // namespace Core

#endif //CORE_PROGRESSINDICATOR_HH
