/** Copyright (C) 2008 Philippe Dreuw
 * 
 * @file    ffmpeg_use.cc
 * @author  Philippe Dreuw <dreuw@cs.rwth-aachen.de>
 * @brief   main executable to exract frames from a video file.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 * 
 * This sample program shows how to use ffmpeg to read a
 * videofile. This version is for the SVN release of ffmpeg.  Use
 * 'make' to build (assuming libavformat, libavcodec, libavutil and
 * libargtable2 are correctly installed on your system).
 *
 * Run using 'ffmpeg_use myvideofile.mpg -s 0.0 --frames=5' to write
 * the first five frames from "myvideofile.mpg" to disk in PNG format.
 * 
 * $Id: ffmpeg_use.cc,v 1.13 2009-11-18 15:42:04 dreuw Exp $
 * 
 */

#define __STDC_CONSTANT_MACROS   // required for e.g. INT64_C macros; must be defined before inclusion of stdint.h and ffmpefile!
#include <stdint.h>

extern "C" {
#include "ffmpegfile.h"
}

#include "ProgressIndicator.hh"
#include "Functor.hh"
#include "Image.hh"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <limits.h>
#include <argtable2.h>

using namespace Core;
using namespace std;


void usage(void **argtable, const char *progname) {
    FILE *fp = stdout;

    fprintf(fp, "Usage: %s ", progname);
    arg_print_syntaxv(fp, argtable, "\n");
    arg_print_glossary(fp, argtable, " %-50s %s\n");
    exit(0);
}

void version(const char *progname) {
    FILE *fp = stdout;

    fprintf(fp, "%s -- %s\nwritten by Philippe Dreuw, 2008 \n", progname, "$Revision: 1.13 $");
    ffmpeg_version();
    exit(0);
}

int main(int argc, char *argv[])
{
    AVFormatContext *pFormatCtx=NULL;
    AVFrame         *pFrame=NULL; 
    AVFrame         *pFrameRGB=NULL;
    uint8_t         *buffer=NULL;
    int             videoStream;
    unsigned int    height, width, streams, colorChannelCount, totalSampleCount, framesPerSecond=25;
    float           fps, duration;
    double          input_start_time, input_end_time;
    double          frame_start_time, frame_end_time;
    int64_t         start_time = AV_NOPTS_VALUE, end_time = AV_NOPTS_VALUE;

    AVFormatParameters *formatParams=NULL;  
    AVFormatParameters formatParamsDevice;
    AVInputFormat *iformat=NULL;

    int nOfFrames, nerrors;
    string filename;
    const char *progname = "ffmpeg_use";

    struct arg_dbl *startTime  = arg_dbl0("s", "startTime", "<time>", "start time in seconds");
    struct arg_dbl *endTime    = arg_dbl0("e", "endTime", "<time>", "end time in seconds");
    struct arg_int *startFrame = arg_int0("i", "startFrame", "<frame>", "start frame as index");
    struct arg_int *endFrame   = arg_int0("j", "endFrame", "<frame>", "end frame as index");
    struct arg_int *frames     = arg_int0("f", "frames","<n>", "nofimages to extract, 0 means until end (default 10)");
    struct arg_str *type       = arg_str0("t", "type","<png|ppm>", "output type of extracted image (default png)");
    struct arg_lit *help       = arg_lit0("h", "help", "show help and exit");
    struct arg_lit *info       = arg_lit0("I", "Info", "show file Info and exit");
    struct arg_lit *versio     = arg_lit0("V", "version", "show version and exit");
    struct arg_lit *verb       = arg_lit0("v", "verbose", "verbose output");
    struct arg_file *file      = arg_file1(NULL,NULL,"<file>","input video file (supported formats: MPEG1/2/4, DivX3/4/5, XviD, WMV7, H.263)");
    struct arg_end *end        = arg_end(20);
    void *argtable[]           = {startTime, endTime, startFrame, endFrame, frames, type, help, info, versio, verb, file, end};

    // --------------------------------------------------------------------- //
    // Setting default values
    // --------------------------------------------------------------------- //   
        
    verb->count         = 0;           // not verbose
    startTime->dval[0]  = 0.0;         // start time
    endTime->dval[0]    = DBL_MAX;     // end time
    frames->ival[0]     = 10;          // extract 10 images
    startFrame->ival[0] = 0;           // start frame
    endFrame->ival[0]   = INT_MAX;     // end frame
    type->sval[0]       = "png";       // default output format

    // --------------------------------------------------------------------- //
    // Parse commandline using libargtable 
    // --------------------------------------------------------------------- //

    if (arg_nullcheck(argtable) != 0)
        printf("error: insufficient memory\n");
    nerrors = arg_parse(argc,argv,argtable);

    // first handle special flags
    if(help->count > 0) {
        usage(argtable, progname);
    }
    if(versio->count > 0) {
        version(progname);
    }
    // then check for errors
    if(nerrors > 0) {
        arg_print_errors(stderr, end, progname);
        fprintf(stderr, "Try '%s --help' for more information.\n",progname);
        exit(EXIT_FAILURE);
    }

    // instantiate objects of PngImage and PpmImage
    Image::PngImage pngImage;
    Image::PpmImage ppmImage;
    
    // instantiate TSpecificFunctor objects
    TSpecificFunctor<Image::PngImage> savePng(&pngImage, &Image::PngImage::save);
    TSpecificFunctor<Image::PpmImage> savePpm(&ppmImage, &Image::PpmImage::save);

    // create TFunctor pointers
    TFunctor *save;
    if (strcmp(type->sval[0], "png") == 0) {
        save = &savePng;
    } else if (strcmp(type->sval[0], "ppm") == 0) {
        save = &savePpm;
    } else {
        printf("ERROR: wrong file type '%s'\n", type->sval[0]);        
        fprintf(stderr, "Try '%s --help' for more information.\n",progname);
        exit(EXIT_FAILURE);
    }
      
    // --------------------------------------------------------------------- //
    // Check parsed arguments
    // --------------------------------------------------------------------- //

    //video4linux device support not yet checked
    if(strstr(file->filename[0], "/dev/") != NULL) {
        printf("WARNING: I've never checked if this realy works :-/\n");
        printf("using video4linux input device '%s'\n", file->filename[0]);
#if defined FFMPEG_VERSION && FFMPEG_VERSION >= 50
        //deprecated: formatParamsDevice.device = file->filename[0]; // e.g. "/dev/video0";
#else
        formatParamsDevice.device = file->filename[0]; // e.g. "/dev/video0";
#endif
        formatParamsDevice.channel = 0;
        formatParamsDevice.standard = "pal";
        //formatParamsDevice.standard = NULL;
        formatParamsDevice.width = 640;
        formatParamsDevice.height = 480;
        iformat = av_find_input_format("video4linux");
        //iformat = av_find_input_format("video4linux2");
        formatParamsDevice.time_base.den = framesPerSecond;
        formatParamsDevice.time_base.num = 1;

        formatParams = &formatParamsDevice;
    }
    filename         = file->filename[0];
    nOfFrames        = frames->ival[0];

    input_start_time = startTime->dval[0]; 
    input_start_time *= AV_TIME_BASE;
    start_time       = (int64_t)input_start_time;
    
    input_end_time   = endTime->dval[0]; 
    input_end_time   *= AV_TIME_BASE;
    end_time         = (int64_t)input_end_time;
    
    // --------------------------------------------------------------------- //
    // start video processing
    // --------------------------------------------------------------------- //
    // Register all formats and codecs
    avcodec_init();
    av_register_all();

#if defined FFMPEG_VERSION && FFMPEG_VERSION >= 50
    // Register all devices
    avdevice_register_all(); 
#endif

    if(ffmpeg_open(&pFormatCtx, filename.c_str(), iformat, &videoStream, formatParams) !=0) {
        exit(EXIT_FAILURE);
    }
    // Allocate memory
    if(ffmpeg_alloc(pFormatCtx, videoStream, &pFrame, &pFrameRGB, &buffer) !=0) {
        printf("couldn't allocate memory\n");
        exit(EXIT_FAILURE);        
    }
    // Get video info
    ffmpeg_read_info(pFormatCtx, videoStream, &height, &width, &fps, &streams, &totalSampleCount, &colorChannelCount, &duration);

    // if start or end _frame_ was specified, we can calculate the start and end
    // _time_ now using fps
    if(startFrame->ival[0] != 0) {
        frame_start_time = float(startFrame->ival[0])/fps;
        frame_start_time *= AV_TIME_BASE;
        start_time       = (int64_t)frame_start_time;
        if(verb->count > 0) {
            printf("changing start time to %lld\n", start_time);        
        }
    }
    if(endFrame->ival[0] != INT_MAX) {
        frame_end_time = float(endFrame->ival[0])/fps;
        frame_end_time *= AV_TIME_BASE;
        end_time       = (int64_t)frame_end_time;
        if(verb->count > 0) {
            printf("changing end time to %lld\n", end_time);        
        }
    }
    
    // If end_time was set, ignore frames
    if(end_time != (int64_t)AV_NOPTS_VALUE) {      
        if(verb->count > 0) {
            printf("calculating the number of frames to extract between start_time = %lld and end_time = %lld\n", start_time, end_time);
        }
        nOfFrames = (int)(( float(end_time - start_time) * fps)/float(AV_TIME_BASE));
        if(nOfFrames < 1) {
            fprintf(stderr, "warning: specified time interval too short to read at least one image.\n");
            fprintf(stderr, "warning: using default value (10).\n");
            nOfFrames=10;
        }        
        else {
            if(verb->count > 0) {
                printf("nOfFrames to extract = %d, AV_TIME_BASE = %d\n", nOfFrames, AV_TIME_BASE);
            }
        }
    } 
    // if nOfFrames is zero then extract all images until the end
    else if (nOfFrames == 0) {
        nOfFrames = (int)(totalSampleCount - (unsigned int)(( float(start_time) * fps)/float(AV_TIME_BASE)));        
    }

    // If seeking (forward/backward possible) requested, we execute it
    if (start_time != (int)AV_NOPTS_VALUE) {
        if(verb->count > 0) {
            printf("seeking to %lld\n", start_time);
        }
        int seek_flags = AVSEEK_FLAG_BACKWARD;
        //int seek_flags = start_time < 0 ? AVSEEK_FLAG_BACKWARD : 0; // 0 or AVSEEK_FLAG_ANY
        // possible seek flags are:
        // AVSEEK_FLAG_BACKWARD (1) (recommended), seek backward
        // AVSEEK_FLAG_BYTE (2)  seeking based on position in bytes
        // AVSEEK_FLAG_ANY (4), seek to any frame, even non keyframes
        if(ffmpeg_seek(pFormatCtx, &start_time, videoStream, seek_flags) != 0) {
            exit(EXIT_FAILURE);
        }                            
        // ffmpeg seeks to the _closest_ keyframe only! if exact frame
        // seeking is wanted, we have to decode one frame and check
        // its position. This is automatically handled in
        // ffmpeg_read_next_frame() by specifying the rescaled
        // start_time.
    }

    // if dump info was required, we will exit here:
    if((verb->count > 0) || (info->count > 0)){
        ffmpeg_dump_info(pFormatCtx, filename.c_str());
        printf("  Input Start Time:  %ld\n", (int64_t)input_start_time);
        printf("  Input Start Pos:   %ld\n", start_time);
        printf("  Number of Frames:  %ld\n", nOfFrames);
        printf("  Full Duration:     %f\n", duration);
        printf("  Total Frames:      %d\n", totalSampleCount);
        if(info->count > 0) exit(0);
    }


    Core::WithProgressIndicator progressIndicator("extracting "+ filename, "frames");
    progressIndicator.start(nOfFrames);
   
    // Read frames and save frames to disk
    if(verb->count > 0) {
        printf("extracting %d image(s)\n", nOfFrames);
    }

    // --------------------------------
    // THE MAIN FRAME EXTRACTION LOOP
    int i=0;   
    while(ffmpeg_read_next_frame(pFormatCtx, videoStream, pFrame, pFrameRGB, start_time) && i<nOfFrames)
        {           
            (*save)(pFrameRGB, width, height, i++);
            progressIndicator.notify(i);
        }
    progressIndicator.finish(false);
    if(i != nOfFrames) {
        fprintf(stderr, "error: could only read %d frames instead of %d\n", i, nOfFrames);
        exit(EXIT_FAILURE);
    }           
    printf("\n");
    // --------------------------------
    
    // Free allocated buffers
    ffmpeg_free(pFrame, pFrameRGB, buffer);
    
    // Close the video file and the codec
    ffmpeg_close(pFormatCtx, videoStream);

    // Free argtable parser
    arg_freetable(argtable,sizeof(argtable)/sizeof(argtable[0]));

    return 0;
}
