/** Copyright (C) 2008 Philippe Dreuw
 * 
 * @file    ffmpegfile.h
 * @author  Philippe Dreuw <dreuw@cs.rwth-aachen.de> 
 * @brief   functions to access video files using the FFmpeg library.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * $Id: ffmpegfile.c,v 1.22 2009-11-18 15:42:05 dreuw Exp $
 * 
 */
#include "ffmpegfile.h" 

#include <inttypes.h> //defines e.g. "PRId64" macros
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined FFMPEG_VERSION && FFMPEG_VERSION >= 50
#warning The <libavutil/pixfmt.h> file has to be considered an internal but installed header, so it should not be directly included in your projects.
#include <libavutil/pixfmt.h>  
#endif

int ffmpeg_open(AVFormatContext **pFormatCtx,
                const char *filename,
                AVInputFormat *iformat,
                int *videoStream,
                AVFormatParameters *formatParams)
{
    AVCodecContext *pCodecCtx=NULL;
    AVCodec        *pCodec=NULL;
    int err;

#ifdef DEBUG
    printf("*** %s:%d debug: ffmpeg_open filename=%s\n",
           __FILE__, __LINE__, filename); 
#endif
    

#if defined FFMPEG_VERSION && FFMPEG_VERSION >= 50
    // Open video file and allocate all the structures needed to read an input stream.
    // Does not open the codec
    if (strcmp(filename, "") != 0) {
        err = av_open_input_file(&(*pFormatCtx), 
                                 filename, 
                                 iformat, 
                                 0, 
                                 formatParams);
        if(err < 0) {
            fprintf(stderr, "error in file '%s' at line %d: couldn't open file '%s'\n",
                    __FILE__, __LINE__, filename);            
            return err;
        }
    }       
#else
    // Open video file and allocate all the structures needed to read an input stream.
    // Does not open the codec    
    if(strstr(filename, "/dev/") != NULL) {
        //video4linux device
        err = av_open_input_stream(&(*pFormatCtx), 
                                   NULL, 
                                   formatParams->device, 
                                   iformat, 
                                   formatParams); 
        if(err < 0) {
            fprintf(stderr, "error in file '%s' at line %d: couldn't open device '%s'\n", 
                    __FILE__, __LINE__, (*formatParams).device);
            return err;
        }
    } else {
        //videofile
        err = av_open_input_file(&(*pFormatCtx), 
                                 filename, 
                                 NULL, 
                                 0, 
                                 NULL);
        if(err < 0) {
            fprintf(stderr, "error in file '%s' at line %d: couldn't open file '%s'\n",
                    __FILE__, __LINE__, filename);            
            return err;
        }
    }
#endif
    
    // Retrieve stream information
    if(av_find_stream_info(*pFormatCtx)<0) {
        fprintf(stderr, "error in file '%s' at line %d: couldn't find stream information\n", 
                __FILE__, __LINE__);
        return -1;
    }
           
#ifdef DEBUG
    // Dump information about file onto standard error
    ffmpeg_dump_info((*pFormatCtx), filename);
#endif

    // Find the first video stream
    (*videoStream) = av_find_default_stream_index((*pFormatCtx));
    if((*videoStream)==-1) {
        fprintf(stderr, "error in file '%s' at line %d: couldn't find a video stream\n", 
                __FILE__, __LINE__);
        return -1; // Didn't find a video stream
    }

    // Get a pointer to the codec context for the video stream
    pCodecCtx = (*pFormatCtx)->streams[*videoStream]->codec;

    // Find the decoder for the video stream
    pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
    if(pCodec==NULL) {
        fprintf(stderr, "error in file '%s' at line %d: couldn't find a codec\n", 
                __FILE__, __LINE__);
        return -1;
    }

    // Open the codec
    if(avcodec_open(pCodecCtx, pCodec)<0) {
        fprintf(stderr, "error in file '%s' at line %d: couldn't open codec\n", 
                __FILE__, __LINE__);
        return -1;
    }
    //free(pCodecCtx);
    return 0;
}

int ffmpeg_alloc(AVFormatContext *pFormatCtx, 
                 int videoStream,
                 AVFrame **pFrame, 
                 AVFrame **pFrameRGB,
                 uint8_t **buffer) 
{
    AVCodecContext *pCodecCtx=NULL;
    int numBytes;

    // Get a pointer to the codec context for the video stream
    pCodecCtx = (pFormatCtx)->streams[videoStream]->codec;

    // Allocate video frame
    (*pFrame)=avcodec_alloc_frame();

    // Allocate an AVFrame structure
    (*pFrameRGB)=avcodec_alloc_frame();
    if((pFrameRGB)==NULL) {
        fprintf(stderr, "error in file '%s' at line %d: couldn't allocate an AVFrame structure\n", 
                __FILE__, __LINE__);
        return -1;
    }

    // Determine required buffer size and allocate buffer
    numBytes=avpicture_get_size(PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);

    (*buffer) = (uint8_t *)malloc(numBytes * sizeof(uint8_t));  //@todo: where will it be freed?

    // Assign appropriate parts of buffer to image planes in pFrameRGB
    avpicture_fill((AVPicture *) (*pFrameRGB), (*buffer), PIX_FMT_RGB24, pCodecCtx->width, pCodecCtx->height);
    return 0;
}

void ffmpeg_version() {
    // output build and version numbers
    fprintf(stderr, "  libavutil version:   %s\n", AV_STRINGIFY(LIBAVUTIL_VERSION));
    fprintf(stderr, "  libavutil build:     %d\n", LIBAVUTIL_BUILD);
    fprintf(stderr, "  libavcodec version:  %s\n", AV_STRINGIFY(LIBAVCODEC_VERSION));
    fprintf(stdout, "  libavcodec build:    %d\n", LIBAVCODEC_BUILD);
    fprintf(stderr, "  libavformat version: %s\n", AV_STRINGIFY(LIBAVFORMAT_VERSION));
    fprintf(stderr, "  libavformat build:   %d\n", LIBAVFORMAT_BUILD);
    fprintf(stderr, "  built on " __DATE__ " " __TIME__);
#ifdef __GNUC__
    fprintf(stderr, ", gcc: " __VERSION__ "\n");
#else
    fprintf(stderr, ", using a non-gcc compiler\n");
#endif
}

void ffmpeg_dump_info(AVFormatContext *pFormatCtx,
                      const char *filename)
{                    
    // Dump information about file onto standard error
    dump_format(pFormatCtx, 0, filename, 0);
}

void ffmpeg_read_info(AVFormatContext *pFormatCtx,
                      int videoStream, 
                      unsigned int *height, 
                      unsigned int *width,
                      float *fps,
                      unsigned int *streams,
                      unsigned int *totalSampleCount,
                      unsigned int *colorChannelCount,
                      float *duration)
{
    AVCodecContext *pCodecCtx=NULL;
    int frate=0, fbase=0;

    pCodecCtx = (pFormatCtx)->streams[videoStream]->codec;
    frate     = pCodecCtx->time_base.den;
    fbase     = pCodecCtx->time_base.num;

    *height  = pCodecCtx->height;
    *width   = pCodecCtx->width;    
    *streams = pFormatCtx->nb_streams;   


#if LIBAVCODEC_BUILD > ((47<<16)+(53<<8)+0)
    *fps = av_q2d ((pFormatCtx)->streams[videoStream]->r_frame_rate);
#else
    *fps = (double)pCodecCtx.frame_rate / (double)pCodecCtx.frame_rate_base;
#endif
    *duration=((float)pFormatCtx->duration/(float)AV_TIME_BASE);                      //exact duration
    //*totalSampleCount=(unsigned int)(ceil((*duration) * (*fps)));                   //legacy - only for boston104 in combination with exact duration

    *duration = (float)((unsigned int)((*duration) * FRAME_PRECISION)/(float)FRAME_PRECISION);  //exact up to a certain precision
    *totalSampleCount=(unsigned int)(floor((*duration) * (*fps)));                            //correct

#ifdef DEBUG
    printf("*** %s:%d debug: height=%d, width=%d, frate=%d, fbase=%d, fps=%f, streams=%d\n", 
           __FILE__, __LINE__,
           *height, *width, frate, fbase, *fps, *streams); 
    printf("*** %s:%d debug: (float)AV_TIME_BASE=%lf, exact duration=%0"PRId64", precision duration=%lf, samples=%d \n",
           __FILE__, __LINE__,
           (float)AV_TIME_BASE, pFormatCtx->duration, *duration, *totalSampleCount);
#endif

    switch (pCodecCtx->pix_fmt) {
    case PIX_FMT_GRAY8:
    case PIX_FMT_MONOWHITE:
    case PIX_FMT_MONOBLACK:
    case PIX_FMT_PAL8:
        *colorChannelCount=1;
        break;
    case PIX_FMT_YUV420P:
    case PIX_FMT_RGB24:
    case PIX_FMT_BGR24:
    case PIX_FMT_YUV422P:
    case PIX_FMT_YUV444P:
    case PIX_FMT_YUV410P:
    case PIX_FMT_YUV411P:
#if defined FFMPEG_VERSION && FFMPEG_VERSION < 50
    case PIX_FMT_YUV422:
    case PIX_FMT_RGBA32:
#endif
        *colorChannelCount=3;
        break;
    default:
        fprintf(stderr, "(%s:%d) warning: unhandled colorModel '%d', assuming 3 channel color model", 
                __FILE__, __LINE__,
                pCodecCtx->pix_fmt);
        *colorChannelCount=3;
    }
#ifdef DEBUG
    printf("*** %s:%d debug: colorChannelCount=%d \n",
           __FILE__, __LINE__,
           *colorChannelCount); 
#endif
}

int ffmpeg_seek(AVFormatContext *pFormatCtx, 
                int64_t *seek_pos,
                int videoStream,
                int seek_flags) 
{
    AVStream *st=NULL;
    int ret, index;
    int64_t old_seek_pos;

#ifdef DEBUG
    printf("*** %s:%d debug: seek_pos=%Ld, videoStream=%d, seek_flags=%d\n",
           __FILE__, __LINE__,
           *seek_pos, videoStream, seek_flags); 
#endif
        
    /* add the stream start time */
    if (pFormatCtx->start_time != (int)AV_NOPTS_VALUE) {
        *seek_pos += pFormatCtx->start_time;
    }

    /* timestamp must be expressed in AV_TIME_BASE units for non-default stream */
    if(videoStream>=0) {  
        st=pFormatCtx->streams[videoStream];
        old_seek_pos=*seek_pos;         
        *seek_pos = av_rescale_q(*seek_pos, AV_TIME_BASE_Q, st->time_base);
#ifdef DEBUG
        printf("*** %s:%d debug: rescaling old_seek_pos=%"PRId64" to new seek_pos=%"PRId64" in AV_TIME_BASE units\n",
               __FILE__, __LINE__,
               old_seek_pos, *seek_pos); 
#endif
    }

    /* searches the index (binary search) which corresponds to the wanted timestamp of seek_pos */
    index = av_index_search_timestamp(st, *seek_pos, seek_flags);
    if (index < 0) {
        fprintf(stderr, "error in file '%s' at line %d: error while seeking to seek_pos=%"PRId64" using seek_flags=%d in file '%s'\n", 
                __FILE__, __LINE__, *seek_pos, seek_flags, pFormatCtx->filename);
        return index;
    }        
    
#ifdef DEBUG
    printf("*** %s:%d debug: seeking to seek_pos=%"PRId64", stream start time=%"PRId64", seek_flags=%d, search index=%d\n",
           __FILE__, __LINE__,
           *seek_pos, pFormatCtx->start_time, seek_flags, index); 
#endif

    ret = av_seek_frame(pFormatCtx, videoStream, *seek_pos, seek_flags); 
    if (ret < 0) {
        fprintf(stderr, "error in file '%s' at line %d: could not seek to position %0.3f in file %s\n", 
                __FILE__, __LINE__,  
                (double)*seek_pos / AV_TIME_BASE, pFormatCtx->filename);
        return ret;
    }
    return 0;
}

int ffmpeg_read_next_frame(AVFormatContext *pFormatCtx, 
                           int videoStream, 
                           AVFrame *pFrame, 
                           AVFrame *pFrameRGB,
                           const int64_t seek_pos)
{
    static struct AVCodecContext  *pCodecCtx=NULL;
    static struct SwsContext *sws_Context=NULL;
    static struct AVPacket packet;
    int           frameFinished=0;
    double pts;
#if defined FFMPEG_VERSION && FFMPEG_VERSION >= 50
    uint8_t *data; 
    int numBytes;
#endif

    //
    // see http://www.dranger.com/ffmpeg/tutorial05.html for more info
    //
    // if we simply synced the video by just counting frames and
    // multiplying by frame rate, there is a chance that it will go out of
    // sync. Packets from the stream might have what is called a decoding
    // time stamp (DTS) and a presentation time stamp (PTS)
    // 
    // Some formats, like MPEG, use what they call "B" frames (B
    // stands for "bidirectional"). The two other kinds of frames are called
    // "I" frames and "P" frames ("I" for "intra" and "P" for "predicted").
    // I frames contain a full image. P frames depend upon previous I
    // and P frames and are like diffs or deltas. B frames are the same as P
    // frames, but depend upon information found in frames that are displayed
    // both before and after them! This explains why we might _not_ have a
    // finished frame after we call avcodec_decode_video.
    //
    // So let's say we had a movie, and the frames were displayed
    // like: I B B P. Now, we need to know the information in P before we can
    // display either B frame. Because of this, the frames might be stored
    // like this: I P B B.
    // The decoding timestamp tells us when we need to decode
    // something, and the presentation time stamp tells us when we
    // need to display something. So, in this case, our stream might look
    // like this:
    // 
    // PTS: 1 4 2 3
    // DTS: 1 2 3 4
    // Stream: I P B B
    //
    // Generally the PTS and DTS will only differ when the stream we
    // are playing has B frames in it.
    //
    // packet.dts := decoding time stamp (DTS) 
    // packet.pts := presentation time stamp (PTS)   
    //   PTS is stored as an integer. This value is a timestamp that
    //   corresponds to a measurement of time in that stream's time_base unit.
    //   We can convert this value to seconds by dividing by the framerate. 
    //   The time_base value of the stream is going to be 1/framerate (for fixed-fps content), 
    //   so to get the PTS in seconds, we multiply by the time_base. 
    //

    // let's start reading frames
    pCodecCtx = (pFormatCtx)->streams[videoStream]->codec;

#ifdef DEBUG
    printf("*** %s:%d debug: reading stream=%d\n",
           __FILE__, __LINE__,
           videoStream);
#endif   
    //av_read_packet allocates memory for packet.data buffer
    while(av_read_frame(pFormatCtx, &packet) >=0)
        {
            // Is this a packet from the video stream?
            if(packet.stream_index==videoStream)
                {
                    pts=0; //init presentation time stamp

                    // Decode next video frame - why we might _not_
                    // have a finished frame, e.g. I/B/P frames in MPEG streams
#if defined FFMPEG_VERSION && FFMPEG_VERSION >= 50
                    avcodec_decode_video2(pCodecCtx, pFrame, &frameFinished, &packet); 
#else
                    avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, 
                                         packet.data, packet.size);
#endif
                    // what we really want is the PTS of our newly
                    // decoded raw frame, so we know when to display it.
                    // ffmpeg reorders the packets so that the DTS of
                    // the packet being processed by avcodec_decode_video() will always be
                    // the same as the PTS of the frame it returns. But, another warning: we
                    // won't always get this information, either.
                    if(packet.dts != AV_NOPTS_VALUE) {
                        pts = packet.dts;
                    } else {
                        pts = 0; //we set the PTS to 0 if we can't figure out what it is. 
                    }
                    //pts *= av_q2d(pFormatCtx->streams[videoStream]->time_base); //if we need seconds instead of index

#ifdef DEBUG
                    printf("*** %s:%d debug: decoding: seek_pos: %"PRId64", dts: %"PRId64", pts: %"PRId64"\n",
                           __FILE__, __LINE__, 
                           seek_pos, 
                           packet.dts,
                           (int64_t)pts);
#endif                  
                    // Did we get a video frame?
                    if(frameFinished)
                        {
                            if ((int64_t)(pts) < seek_pos) {
                                // we got a frame but did not yet decode the right image,
                                // i.e. closest seekable keyframe has been before
                                // wanted start frame -> decode next frame
                                frameFinished=0;
#ifdef DEBUG
                                printf("*** %s:%d debug: decoding: frameFinished: not yet finished the correct frame. pts: %"PRId64" < seek_pos: %"PRId64" \n",
                                       __FILE__, __LINE__, 
                                       pts, seek_pos);
#endif
                                continue;
                            }

                            // we got the right frame!
#ifdef DEBUG
                            printf("*** %s:%d debug: decoding: frame finished. seek_pos: %"PRId64", dts: %"PRId64", pts: %"PRId64"\n",
                                   __FILE__, __LINE__, 
                                   seek_pos, 
                                   packet.dts,
                                   (int64_t)pts);
#endif
                            // Convert the image from its native format to RGB
#if LIBAVCODEC_BUILD >= ((51<<16)+(38<<8)+0)
                            // patched to use sws_scale instead of img_convert:
                            sws_Context = sws_getCachedContext(sws_Context,
                                                               pCodecCtx->width, pCodecCtx->height, pCodecCtx->pix_fmt, /* src format */
                                                               pCodecCtx->width, pCodecCtx->height, PIX_FMT_RGB24,      /* destination format */
                                                               SWS_BICUBIC, NULL, NULL, NULL);
                            if (!sws_Context) {
                                fprintf(stderr, "error in file '%s' at line %d: ffmpeg-conversion failed (%d->%d)\n",
                                        __FILE__, __LINE__, pCodecCtx->pix_fmt, PIX_FMT_RGB24);            
                            }
                            int res = sws_scale(sws_Context,
                                                pFrame->data, pFrame->linesize, 0, pCodecCtx->height, /* src format */
                                                pFrameRGB->data, pFrameRGB->linesize);                /* destination format */
                            if (res<0) {
                                fprintf(stderr, "error in file '%s' at line %d: ffmpeg-conversion failed (%d->%d)\n",
                                        __FILE__, __LINE__, pCodecCtx->pix_fmt, PIX_FMT_RGB24);            
                            }
#else
                            // img_convert is deprecated!
                            img_convert((AVPicture *)pFrameRGB, PIX_FMT_RGB24, 
                                        (AVPicture*)pFrame, pCodecCtx->pix_fmt, pCodecCtx->width,  
                                        pCodecCtx->height);
#endif
                            break; //we got our frame
                        }
                }            
        }
    // Free the packet that was allocated by av_read_frame()
    av_free_packet(&packet);
    return frameFinished;
}
    
void ffmpeg_free(AVFrame *pFrame, 
                 AVFrame *pFrameRGB,
                 uint8_t *buffer) 
{    
    // Free the RGB image
    av_free(buffer);
    av_free(pFrameRGB);
    
    // Free the YUV frame
    av_free(pFrame);

    buffer=NULL;
    pFrameRGB=NULL;
    pFrame=NULL;
}

void ffmpeg_close(AVFormatContext *pFormatCtx, 
                  int videoStream)
{
    // Close the codec and free the packet that was eventually allocated by
    // av_read_frame() function
    avcodec_close((pFormatCtx)->streams[videoStream]->codec);

    // Close the video file and free all allocated structures needed to read the input stream
    av_close_input_file(pFormatCtx);

    pFormatCtx=NULL;
}
