/** Copyright (C) 2008 Philippe Dreuw
 * 
 * @file    ffmpegfile.h
 * @author  Philippe Dreuw <dreuw@cs.rwth-aachen.de>
 * @brief   functions to access video files using the FFmpeg library.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * $Id: ffmpegfile.h,v 1.18 2009-11-18 15:42:06 dreuw Exp $
 *
 */
#ifndef FFMPEG_FILE_H
#define FFMPEG_FILE_H


#ifdef __cplusplus
extern "C" {
#endif


#if defined FFMPEG_VERSION && FFMPEG_VERSION >= 50
#include <libavutil/avutil.h>
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libswscale/swscale.h>
#include <libavdevice/avdevice.h>
#else
#include <ffmpeg/avutil.h>
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>
#include <ffmpeg/swscale.h>
#endif


#define FRAME_PRECISION 10000
    
    /**
     * open a video file or a video device (video4linux) and allocates all necessary memory.
     *
     * @param pFormatCtx    a reference to an AVFormatContext pointer
     * @param filename      the filename of the video that should be opened
     *                      can be a video4linux device
     * @param iformat       a pointer to an AVInputFormat struct if a video4linux device
     *                      should be opened, otherwise NULL
     * @param videoStream   a reference where the videostream id will be stored
     * @param formatParams  a reference to an AVFormatParameters struct defining the 
     *                      the name and resolution of the video4linux device, 
     *                      otherwise NULL or an empty struct
     *
     * @return returns 0 if successful, otherwise -1
     */
    int ffmpeg_open(AVFormatContext **pFormatCtx,
                    const char *filename,
                    AVInputFormat *iformat,
                    int *videoStream,
                    AVFormatParameters *formatParams);
    
    /**
     * allocates all necessary memory to store a frame of a given video stream.
     *
     * @param pFormatCtx    a pointer the AV context
     * @param videoStream   the id of the stream for which memory should be allocated
     * @param pFrame        a reference to an AVFrame pointer, 
     *                      which will store the original AV image (YUV)
     * @param pFrameRGB     pFrameRGB to an AVFrame pointer, 
     *                      which will store the converted RGB image
     * @param buffer        additional reference to a buffer pointer, 
     *                      which will be used to read and convert the AV frames
     */
    int ffmpeg_alloc(AVFormatContext *pFormatCtx, 
                     int videoStream,
                     AVFrame **pFrame, 
                     AVFrame **pFrameRGB,
                     uint8_t **buffer);

    /**
     * dump some useful info about the currently used ffmpeg versions and builds
     * to stdout.
     */
    void ffmpeg_version();

    /**
     * dump some useful info for a given context and filename to stdout.
     */
    void ffmpeg_dump_info(AVFormatContext *pFormatCtx,
                          const char *filename);
    
    /**
     * reads some useful info for a given context and video stream id.
     *
     * @param pFormatCtx         the pointer to the AV context
     * @param videoStream        the index of the video stream from which the info should be extracted
     * @param height             the height of the video will be stored here
     * @param width              the width of the video will be stored here
     * @param fps                the framerate of the video will be stored here
     * @param streams            the total number of streams (audio, video, subtitlem, ...) 
     *                           inside the video will be stored here
     * @param totalSampleCount   the total number of frames will be stored here
     * @param colorChannelCount  will be set to 1 if the detected pixel format is PIX_FMT_GRAY8, 
     *                           will be set to 3 if the detected pixel format is PIX_FMT_YUV420P, 
     *                           otherwise to 3 (default)
     * @param duration           the full length of the video in seconds
     */
    void ffmpeg_read_info(AVFormatContext *pFormatCtx,
                          int videoStream, 
                          unsigned int *height, 
                          unsigned int *width,
                          float *fps, 
                          unsigned int *streams,
                          unsigned int *totalSampleCount,
                          unsigned int *colorChannelCount,
                          float *duration);
    
    /**
     * seeks to a given start time in the video.
     *
     * @param seek_pos           a long integer value representing the start time in seconds 
     *                           multiplicated with the AV time base AV_TIME_BASE 
     *                           will be rescaled depending on stream's codec
     * @param videoStream        if -1 a default stream will be searched, 
     *                           otherwise the given one will be used
     * @param seek_flag          possible seek flags are  AVSEEK_FLAG_BACKWARD(1), AVSEEK_FLAG_BYTE(2), or AVSEEK_FLAG_ANY(4).
     *                           if AVSEEK_FLAG_BACKWARD then seek_pos will correspond to 
     *                           the timestamp which is <= the requested one
     *                           if AVSEEK_FLAG_ANY seek to any keyframe,
     *                           if (0), then only frames >= the requested one otherwise
     * @return < 0 if no such timestamp could be found
     */
    int ffmpeg_seek(AVFormatContext *pFormatCtx,  
                    int64_t *seek_pos,
                    int videoStream,
                    int seek_flag);

    /**
     * reads the next frame from the current stream position and fowards it by
     * one frame.
     *
     * @param pFormatCtx     a pointer to the AV context
     * @param videoStream    the id of the video stream from which the next frame should be read
     * @param pFrame         a pointer to an already allocated AV frame 
     *                       where the the original YUV frame will be stored
     * @param pFrameRGB      a pointer to an already allocated AV frame 
     *                       where the the converted RGB frame will be stored
     * @param seek_pos       if set, the decoded frame position is compared to this position; 
     *                       if lower, then we did not yet decode the right frame, 
     *                       i.e. the closest seekable keyframe has been before the 
     *                       wanted start frame
     * @return               returns 1 if a frame was finished (decoded), otherwise 0
     */
    int ffmpeg_read_next_frame(AVFormatContext *pFormatCtx, 
                               int videoStream, 
                               AVFrame *pFrame, 
                               AVFrame *pFrameRGB,
                               const int64_t seek_pos);

    /**
     * frees the memory allocated in ffmpeg_alloc().
     */
    void ffmpeg_free(AVFrame *pFrame, 
                     AVFrame *pFrameRGB,
                     uint8_t *buffer);

    /**
     * closes the video and the used codec(s).
     *
     * @param pFormatCtx     a pointer to the AV context
     * @param videoStream    the id of the video stream for which a codec was opened
     */
    void ffmpeg_close(AVFormatContext *pFormatCtx, 
                      int videoStream);
    
    
#ifdef __cplusplus
}		/* extern "C" */
#endif	/* __cplusplus */


#endif  /* FFMPEG_FILE_H */
