/** Copyright (C) 2008 Thomas Deselaers
 * 
 * @file    pngfile.c
 * @author  Thomas Deselaers <deselaers@i6.informatik.rwth-aachen.de>
 * @brief   functions to access png files using the png library.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * 
 * $Id: pngfile.c,v 1.2 2008-02-25 15:54:28 dreuw Exp $
 * 
 */
#ifdef __cplusplus
extern "C" {
#endif	/* __cplusplus */

#include "pngfile.h"

#include <png.h>    // libpng only supports C
#include <stdio.h>
#include <stdlib.h>

/* Read INFO PNG file. */
  int read_png_info(const char *file_name, unsigned int *xdim, unsigned int *ydim, unsigned int *ctype) {  /* We need to open the file */
    png_structp png_ptr;
    png_infop info_ptr;
    unsigned int sig_read = 0;
    png_uint_32 width, height;
    int bit_depth, color_type, interlace_type;
    FILE *fp;
    
    if ((fp = fopen(file_name, "rb")) == NULL) {
        //DBGI(DBG_ERROR,printf("Could not open PNG file %s \n", file_name));
        printf("Could not open PNG file %s \n", file_name);
      return -1;
    }
    
    /* Create and initialize the png_struct with the desired error handler
     * functions.  If you want to use the default stderr and longjump method,
     * you can supply NULL for the last three parameters.  We also supply the
     * the compiler header file version, so that we know if the application
     * was compiled with a compatible version of the library.  REQUIRED
     */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    
    if (png_ptr == NULL)
      {
	fclose(fp);
	//DBGI(DBG_ERROR, printf("Error creating PNG struct \n"));
	printf("Error creating PNG struct \n");
	return -1;
      }
    
    /* Allocate/initialize the memory for image information.  REQUIRED. */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL)
      {
	fclose(fp);
	png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
	//DBGI(DBG_ERROR,printf("Error creating PNG info struct \n"));
	printf("Error creating PNG info struct \n");
	return -1;
      }
    
    /* Set error handling if you are using the setjmp/longjmp method (this is
     * the normal method of doing things with libpng).  REQUIRED unless you
     * set up your own error handlers in the png_create_read_struct() earlier.
     */
    if (setjmp(png_ptr->jmpbuf))
      {
	/* Free all of the memory associated with the png_ptr and info_ptr */
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
	fclose(fp);
	/* If we get here, we had a problem reading the file */
	//DBGI(DBG_ERROR,printf("we had a problem reading the file %s \n", file_name));
	printf("we had a problem reading the file %s \n", file_name);
	return -1;
      }
    
    
    /* Set up the input control if you are using standard C streams */
    png_init_io(png_ptr, fp);
    
    /* If we have already read some of the signature */
    png_set_sig_bytes(png_ptr, sig_read);
    
    /* The call to png_read_info() gives us all of the information from the
     * PNG file before the first IDAT (image data chunk).  REQUIRED
     */
    png_read_info(png_ptr, info_ptr);
    
    png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
		 &interlace_type, NULL, NULL);
    
    *xdim  = width;
    *ydim  = height;
    *ctype = color_type;

    /* verbose output */
    //DBGI(DBG_VERBOSE,printf("PNG Info \n"));
    //printf("PNG Info \n");
    //DBGI(DBG_VERBOSE,printf("width = %ld, height = %ld, bit_depth = %d, color_type = %d, interlace_type = %d \n",width,height,bit_depth,color_type,interlace_type));
    //printf("width = %ld, height = %ld, bit_depth = %d, color_type = %d, interlace_type = %d \n",width,height,bit_depth,color_type,interlace_type);
    if (color_type == PNG_COLOR_TYPE_RGB) {
        //DBGI(DBG_VERBOSE,printf("RGB \n"));
        //printf("RGB \n");
    }
    
    /* clean up after the read, and free any memory allocated - REQUIRED */
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    
    /* close the file */
    fclose(fp); 
    return 0;
  }




  /* Read a PNG file. */
  void read_png(const char *file_name, png_byte **row_pointers, unsigned int *xdim, unsigned int *ydim) {  /* We need to open the file */
    png_structp png_ptr;
    png_infop info_ptr;
    unsigned int sig_read = 0;
    png_uint_32 width, height, row;
    int bit_depth, color_type, interlace_type;
    FILE *fp;
    
    if ((fp = fopen(file_name, "rb")) == NULL) {
      //DBGI(DBG_ERROR,printf("Could not open PNG file %s \n", file_name));
      printf("Could not open PNG file %s \n", file_name);
      return;
    }
    
    /* Create and initialize the png_struct with the desired error handler
     * functions.  If you want to use the default stderr and longjump method,
     * you can supply NULL for the last three parameters.  We also supply the
     * the compiler header file version, so that we know if the application
     * was compiled with a compatible version of the library.  REQUIRED
     */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    
    if (png_ptr == NULL)
      {
	fclose(fp);
	//DBGI(DBG_ERROR,printf("Error creating PNG struct \n"));
	printf("Error creating PNG struct \n");
	return;
      }
    
    /* Allocate/initialize the memory for image information.  REQUIRED. */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL)
      {
	fclose(fp);
	png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
	//DBGI(DBG_ERROR,printf("Error creating PNG info struct \n"));
	printf("Error creating PNG info struct \n");
	return;
      }
    
    /* Set error handling if you are using the setjmp/longjmp method (this is
     * the normal method of doing things with libpng).  REQUIRED unless you
     * set up your own error handlers in the png_create_read_struct() earlier.
     */
    if (setjmp(png_ptr->jmpbuf))
      {
	/* Free all of the memory associated with the png_ptr and info_ptr */
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
	fclose(fp);
	/* If we get here, we had a problem reading the file */
	//DBGI(DBG_ERROR,printf("we had a problem reading the file %s \n", file_name));
	printf("we had a problem reading the file %s \n", file_name);
      return;
      }


    /* Set up the input control if you are using standard C streams */
    png_init_io(png_ptr, fp);

    /* If we have already read some of the signature */
    png_set_sig_bytes(png_ptr, sig_read);
    
    /* The call to png_read_info() gives us all of the information from the
     * PNG file before the first IDAT (image data chunk).  REQUIRED
     */
    png_read_info(png_ptr, info_ptr);
    
    png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
		 &interlace_type, NULL, NULL);

    
    /**** Set up the data transformations you want.  Note that these are all
     **** optional.  Only call them if you want/need them.  Many of the
     **** transformations only work on specific types of images, and many
     **** are mutually exclusive.
     ****/
    
    /* tell libpng to strip 16 bit/color files down to 8 bits/color */
    if (bit_depth == 16)
      png_set_strip_16(png_ptr);
    
    /* Expand paletted colors into true RGB triplets */
    if (color_type == PNG_COLOR_TYPE_PALETTE)
      png_set_expand(png_ptr);
    
    /* Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel */
    if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
      png_set_expand(png_ptr);

    /* swap bytes of 16 bit files to least significant byte first */
    if (bit_depth == 16)
      png_set_swap(png_ptr);


    /* After setting the transformations, libpng can update your png_info structure to reflect any transformations you've requested with this call. */
    png_read_update_info(png_ptr, info_ptr);

    /* Allocate the memory to hold the image using the fields of info_ptr. */

    /* The easiest way to read the image: */
    
    for (row = 0; row < height; row++)
      {
	row_pointers[row] = (png_bytep) malloc(png_get_rowbytes(png_ptr, info_ptr));
      }

    /* Now it's time to read the image. Read the entire image in one go */
    png_read_image(png_ptr, row_pointers);
    
    /* read rest of file, and get additional chunks in info_ptr - REQUIRED */
    png_read_end(png_ptr, info_ptr);
    
    /* clean up after the read, and free any memory allocated - REQUIRED */
     png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    
    /* close the file */
    fclose(fp); 
    
  }




  /* write a png file */
  void write_png(const char *file_name, png_byte *image, unsigned int width, unsigned int height, unsigned int bit_depth, unsigned int SamplesPerPixel) {
    FILE *fp;
    png_structp png_ptr;
    png_infop info_ptr;


    /* open the file */
    fp = fopen(file_name, "wb");
    if (fp == NULL)
      return;

    /* Create and initialize the png_struct with the desired error handler
     * functions.  If you want to use the default stderr and longjump method,
     * you can supply NULL for the last three parameters.  We also check that
     * the library version is compatible with the one used at compile time,
     * in case we are using dynamically linked libraries.  REQUIRED.
     */
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    
    if (png_ptr == NULL)
      {
	fclose(fp);
	return;
      }

    /* Allocate/initialize the image information data.  REQUIRED */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL)
      {
	fclose(fp);
	png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
	return;
      }
    
    /* Set error handling.  REQUIRED if you aren't supplying your own
     * error hadnling functions in the png_create_write_struct() call.
     */
    if (setjmp(png_ptr->jmpbuf))
      {
	/* If we get here, we had a problem reading the file */
	fclose(fp);
	png_destroy_write_struct(&png_ptr,  &info_ptr);
	return;
      }
    
    /* set up the output control if you are using standard C streams */
    png_init_io(png_ptr, fp);
    
    
    /* Set the image information here.  Width and height are up to 2^31,
     * bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
     * the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
     * PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
     * or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
     * PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
     * currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
     */
    if(SamplesPerPixel == 1) {
      /* we are dealing with a grayscale image */
      png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth, PNG_COLOR_TYPE_GRAY,
		   PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
    }
    else if(SamplesPerPixel == 3) {
      /* we are dealing with a color image */
      png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth, PNG_COLOR_TYPE_RGB,
		   PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
    }

    /* TODO: PALETTE IMAGES
     * set the palette if there is one.  REQUIRED for indexed-color images 
     */
    /*
      palette = (png_colorp)png_malloc(png_ptr, 256 * sizeof (png_color));
      // ... set palette colors ... 
      png_set_PLTE(png_ptr, info_ptr, palette, 256);
    */
    
  
    /* Optionally write comments into the image 
    text_ptr[0].key = "Title";
    text_ptr[0].text = "Mona Lisa";
    text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
    text_ptr[1].key = "Author";
    text_ptr[1].text = "Leonardo DaVinci";
    text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
    text_ptr[2].key = "Description";
    text_ptr[2].text = "<long text>";
    text_ptr[2].compression = PNG_TEXT_COMPRESSION_zTXt;
    png_set_text(png_ptr, info_ptr, text_ptr, 3);
    */
    
    /* Write the file header information.  REQUIRED */
    png_write_info(png_ptr, info_ptr);

    /* The easiest way to write the image (you may have a different memory
     * layout, however, so choose what fits your needs best).  You need to
     * use the first method if you aren't handling interlacing yourself.
     */
    
    png_uint_32 k;
    png_bytep row_pointers[height];
    for (k = 0; k < height; k++)
      row_pointers[k] = (png_byte *)(image + k*width * SamplesPerPixel);
    
    png_write_image(png_ptr, row_pointers);
    png_write_flush(png_ptr);
    png_write_end(png_ptr, info_ptr);

	png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(fp);
  }




  //free memory
  void free_png(png_bytep *row_pointers, unsigned int height) {
    size_t i;
    for(i=0; i < height; ++i) {
      free(row_pointers[i]);
    }
  }


#if defined __cplusplus
}
#endif


